@@ -8,6 +8,7 @@ import os
 import pathlib
 import platform
 import re
+import shutil
 import sys
 import sysconfig
 from typing import Any, Sequence, cast
@@ -56,6 +57,7 @@ from ruff_lsp import __version__, utils
 USER_DEFAULTS: dict[str, str] = {}
 WORKSPACE_SETTINGS: dict[str, dict[str, Any]] = {}
 INTERPRETER_PATHS: dict[str, str] = {}
+EXECUTABLE_VERSIONS: dict[str, str] = {}
 CLIENT_CAPABILITIES: dict[str, bool] = {
     CODE_ACTION_RESOLVE: True,
 }
@@ -69,7 +71,15 @@ LSP_SERVER = server.LanguageServer(
 
 TOOL_MODULE = "ruff.exe" if platform.system() == "Windows" else "ruff"
 TOOL_DISPLAY = "Ruff"
-TOOL_ARGS = ["--no-cache", "--no-fix", "--quiet", "--format", "json", "-"]
+TOOL_ARGS = [
+    "--force-exclude",
+    "--no-cache",
+    "--no-fix",
+    "--quiet",
+    "--format",
+    "json",
+    "-",
+]
 
 
 ###
@@ -95,7 +105,7 @@ def did_save(params: DidSaveTextDocumentParams) -> None:
 
 @LSP_SERVER.feature(TEXT_DOCUMENT_DID_CHANGE)
 def did_change(params: DidChangeTextDocumentParams) -> None:
-    """LSP handler for textDocument/didSave request."""
+    """LSP handler for textDocument/didChange request."""
     document = LSP_SERVER.workspace.get_document(params.text_document.uri)
     diagnostics: list[Diagnostic] = _linting_helper(document)
     LSP_SERVER.publish_diagnostics(document.uri, diagnostics)
@@ -130,11 +140,14 @@ def _parse_output_using_regex(content: str) -> list[Diagnostic]:
     #   {
     #     "code": "F841",
     #     "message": "Local variable `x` is assigned to but never used",
-    #     "fixed": false,
     #     "location": {
     #       "row": 2,
     #       "column": 5
     #     },
+    #     "end_location": {
+    #       "row": 2,
+    #       "column": 6
+    #     },
     #     "fix": {
     #       "content: "",
     #       "location": {
@@ -146,7 +159,8 @@ def _parse_output_using_regex(content: str) -> list[Diagnostic]:
     #         "column: 0
     #       }
     #     },
-    #     "filename": "/path/to/test.py"
+    #     "filename": "/path/to/test.py",
+    #     "noqa_row": 2
     #   },
     #   ...
     # ]
@@ -162,30 +176,45 @@ def _parse_output_using_regex(content: str) -> list[Diagnostic]:
         diagnostic = Diagnostic(
             range=Range(start=start, end=end),
             message=check.get("message"),
-            severity=_get_severity(check["code"], check.get("type", "Error")),
+            severity=_get_severity(check["code"]),
             code=check["code"],
             source=TOOL_DISPLAY,
-            data=check.get("fix"),
-            tags=(
-                [DiagnosticTag.Unnecessary]
-                if check["code"]
-                in {
-                    "F401",  # `module` imported but unused
-                    "F841",  # local variable `name` is assigned to but never used
-                }
-                else None
+            data=DiagnosticData(
+                fix=check.get("fix"),
+                # Available since Ruff v0.0.253.
+                noqa_row=check.get("noqa_row"),
             ),
+            tags=_get_tags(check["code"]),
         )
         diagnostics.append(diagnostic)
 
     return diagnostics
 
 
-def _get_severity(*_codes: list[str]) -> DiagnosticSeverity:
-    return DiagnosticSeverity.Warning
+def _get_tags(code: str) -> list[DiagnosticTag] | None:
+    if code in {
+        "F401",  # `module` imported but unused
+        "F841",  # local variable `name` is assigned to but never used
+    }:
+        return [DiagnosticTag.Unnecessary]
+    return None
 
 
-NOQA_REGEX = re.compile(r"(?i:# noqa)(?::\s?(?P<codes>([A-Z]+[0-9]+(?:[,\s]+)?)+))?")
+def _get_severity(code: str) -> DiagnosticSeverity:
+    if code in {
+        "F821",  # undefined name `name`
+        "E902",  # `IOError`
+        "E999",  # `SyntaxError`
+    }:
+        return DiagnosticSeverity.Error
+    else:
+        return DiagnosticSeverity.Warning
+
+
+NOQA_REGEX = re.compile(
+    r"(?i:# (?:(?:ruff|flake8): )?(?P<noqa>noqa))"
+    r"(?::\s?(?P<codes>([A-Z]+[0-9]+(?:[,\s]+)?)+))?"
+)
 CODE_REGEX = re.compile(r"[A-Z]{1,3}[0-9]{3}")
 
 
@@ -242,6 +271,11 @@ class Fix(TypedDict):
     end_location: Location
 
 
+class DiagnosticData(TypedDict):
+    fix: Fix | None
+    noqa_row: int | None
+
+
 @LSP_SERVER.feature(
     TEXT_DOCUMENT_CODE_ACTION,
     CodeActionOptions(
@@ -278,7 +312,7 @@ def code_action(params: CodeActionParams) -> list[CodeAction] | None:
                 and len(params.context.only) == 1
                 and kind in params.context.only
             ):
-                results = _formatting_helper(document, select="I001")
+                results = _formatting_helper(document, only="I001")
                 if results is not None:
                     return [
                         CodeAction(
@@ -315,7 +349,8 @@ def code_action(params: CodeActionParams) -> list[CodeAction] | None:
                                 diagnostic
                                 for diagnostic in params.context.diagnostics
                                 if diagnostic.source == "Ruff"
-                                and diagnostic.data is not None
+                                and cast(DiagnosticData, diagnostic.data)["fix"]
+                                is not None
                             ],
                         ),
                     ]
@@ -340,7 +375,7 @@ def code_action(params: CodeActionParams) -> list[CodeAction] | None:
                     ),
                 )
             else:
-                results = _formatting_helper(document, select="I001")
+                results = _formatting_helper(document, only="I001")
                 if results is not None:
                     actions.append(
                         CodeAction(
@@ -380,7 +415,8 @@ def code_action(params: CodeActionParams) -> list[CodeAction] | None:
                                 diagnostic
                                 for diagnostic in params.context.diagnostics
                                 if diagnostic.source == "Ruff"
-                                and diagnostic.data is not None
+                                and cast(DiagnosticData, diagnostic.data)["fix"]
+                                is not None
                             ],
                         ),
                     )
@@ -389,9 +425,8 @@ def code_action(params: CodeActionParams) -> list[CodeAction] | None:
     if not params.context.only or CodeActionKind.QuickFix in params.context.only:
         for diagnostic in params.context.diagnostics:
             if diagnostic.source == "Ruff":
-                if diagnostic.data is not None:
-                    fix = cast(Fix, diagnostic.data)
-
+                fix = cast(DiagnosticData, diagnostic.data)["fix"]
+                if fix is not None:
                     title: str
                     if fix.get("message"):
                         title = f"Ruff ({diagnostic.code}): {fix['message']}"
@@ -405,9 +440,50 @@ def code_action(params: CodeActionParams) -> list[CodeAction] | None:
                             title=title,
                             kind=CodeActionKind.QuickFix,
                             data=params.text_document.uri,
-                            edit=_create_workspace_edit(
-                                document, cast(Fix, diagnostic.data)
-                            ),
+                            edit=_create_workspace_edit(document, fix),
+                            diagnostics=[diagnostic],
+                        ),
+                    )
+
+    # Add "Disable for this line" for every diagnostic.
+    if not params.context.only or CodeActionKind.QuickFix in params.context.only:
+        for diagnostic in params.context.diagnostics:
+            if diagnostic.source == "Ruff":
+                noqa_row = cast(DiagnosticData, diagnostic.data)["noqa_row"]
+                if noqa_row is not None:
+                    line = document.lines[noqa_row - 1].rstrip("\r\n")
+                    match = NOQA_REGEX.search(line)
+                    # `foo  # noqa: OLD` -> `foo  # noqa: OLD,NEW`
+                    if match and match.group("codes") is not None:
+                        codes = match.group("codes") + f", {diagnostic.code}"
+                        start, end = match.start("codes"), match.end("codes")
+                        new_line = line[:start] + codes + line[end:]
+                    # `foo  # noqa` -> `foo  # noqa: NEW`
+                    elif match:
+                        end = match.end("noqa")
+                        new_line = line[:end] + f": {diagnostic.code}" + line[end:]
+                    # `foo` -> `foo  # noqa: NEW`
+                    else:
+                        new_line = f"{line}  # noqa: {diagnostic.code}"
+                    fix = Fix(
+                        content=new_line,
+                        message=None,
+                        location=Location(
+                            row=noqa_row,
+                            column=0,
+                        ),
+                        end_location=Location(
+                            row=noqa_row,
+                            column=len(line),
+                        ),
+                    )
+
+                    actions.append(
+                        CodeAction(
+                            title=f"Ruff: Disable {diagnostic.code} for this line",
+                            kind=CodeActionKind.QuickFix,
+                            data=params.text_document.uri,
+                            edit=_create_workspace_edit(document, fix),
                             diagnostics=[diagnostic],
                         ),
                     )
@@ -429,7 +505,7 @@ def resolve_code_action(params: CodeAction) -> CodeAction:
     ):
         # Generate the "Ruff: Organize Imports" edit
         params.edit = _create_workspace_edits(
-            document, _formatting_helper(document, select="I001") or []
+            document, _formatting_helper(document, only="I001") or []
         )
     elif settings["fixAll"] and params.kind in (
         CodeActionKind.SourceFixAll,
@@ -453,13 +529,26 @@ def apply_autofix(arguments: tuple[TextDocument]):
     )
 
 
+@LSP_SERVER.command("ruff.applyOrganizeImports")
+def apply_organize_imports(arguments: tuple[TextDocument]):
+    uri = arguments[0]["uri"]
+    text_document = LSP_SERVER.workspace.get_document(uri)
+    LSP_SERVER.apply_edit(
+        _create_workspace_edits(
+            text_document, _formatting_helper(text_document, only="I001") or []
+        ),
+        "Ruff: Format imports",
+    )
+
+
 def _formatting_helper(
-    document: workspace.Document, *, select: str | None = None
+    document: workspace.Document, *, only: str | None = None
 ) -> list[TextEdit] | None:
     result = _run_tool_on_document(
         document,
         use_stdin=True,
-        extra_args=["--fix", "--select", select] if select else ["--fix"],
+        extra_args=["--fix"],
+        only=only,
     )
     if result is None:
         return []
@@ -689,6 +778,7 @@ def _get_settings_by_document(document: workspace.Document | None) -> dict[str,
 def _executable_path(settings: dict[str, Any]) -> str:
     """Returns the path to the executable."""
     bundle = get_bundle()
+    print("FOO")
     if settings["path"]:
         # 'path' setting takes priority over everything.
         for path in settings["path"]:
@@ -714,34 +804,47 @@ def _executable_path(settings: dict[str, Any]) -> str:
             )
 
         path = os.path.join(INTERPRETER_PATHS[settings["interpreter"][0]], TOOL_MODULE)
-        if bundle and not os.path.exists(path):
-            log_to_output(
-                f"External interpreter executable ({path}) not found; "
-                f"falling back to bundled executable: {bundle}"
-            )
-            path = bundle
-        else:
-            log_to_output(f"Using external interpreter executable: {path}")
+    else:
+        path = os.path.join(sysconfig.get_path("scripts"), TOOL_MODULE)
+
+    # First choice: the executable in the current interpreter's scripts directory.
+    if os.path.exists(path):
+        log_to_output(f"Using interpreter executable: {path}")
         return path
 
-    # If the interpreter is same as the interpreter running this process, get the
-    # script path directly.
-    path = os.path.join(sysconfig.get_path("scripts"), TOOL_MODULE)
-    if bundle and not os.path.exists(path):
+    # Second choice: the executable in the global environment.
+    environment_path = shutil.which("ruff")
+    if environment_path:
+        log_to_output(f"Using environment executable: {environment_path}")
+        return environment_path
+
+    # Third choice: bundled executable.
+    if bundle:
         log_to_output(
             f"Interpreter executable ({path}) not found; "
             f"falling back to bundled executable: {bundle}"
         )
-        path = bundle
-    else:
-        log_to_output(f"Using interpreter executable: {path}")
+        return bundle
+
+    # Last choice: just return the expected path for the current interpreter.
+    log_to_output(f"Unable to find interpreter executable: {path}")
     return path
 
 
+def _executable_version(executable: str) -> str:
+    """Returns the version of the executable."""
+    if executable not in EXECUTABLE_VERSIONS:
+        version = utils.version(executable)
+        log_to_output(f"Inferred version {version} for: {executable}")
+        EXECUTABLE_VERSIONS[executable] = version
+    return EXECUTABLE_VERSIONS[executable]
+
+
 def _run_tool_on_document(
     document: workspace.Document,
     use_stdin: bool = False,
     extra_args: Sequence[str] = [],
+    only: str | None = None,
 ) -> utils.RunResult | None:
     """Runs tool on the given document.
 
@@ -759,9 +862,17 @@ def _run_tool_on_document(
     # Deep copy, to prevent accidentally updating global settings.
     settings = copy.deepcopy(_get_settings_by_document(document))
 
-    argv: list[str] = (
-        [_executable_path(settings)] + TOOL_ARGS + settings["args"] + list(extra_args)
-    )
+    executable = _executable_path(settings)
+    argv: list[str] = [executable] + TOOL_ARGS + settings["args"] + list(extra_args)
+
+    # If we're trying to run a single rule, add it to the command line, and disable
+    # all other rules (if the Ruff version is sufficiently recent).
+    if only:
+        if _executable_version(executable) >= "0.0.198":
+            argv += ["--extend-ignore", "ALL"]
+        argv += ["--extend-select", only]
+
+    # If we're using stdin, provide the filename.
     if use_stdin:
         argv += ["--stdin-filename", document.path]
     else:
